@@ -835,8 +835,10 @@ impl Session<Playing> {
         }
 
         // Send a new one and reset the timer.
+        // Use a SET_PARAMETER with no body for keepalives, as recommended in the
+        // ONVIF Streaming Specification version version 21.06 section 5.2.2.2.
         let mut req = rtsp_types::Request::builder(
-            rtsp_types::Method::GetParameter,
+            rtsp_types::Method::SetParameter,
             rtsp_types::Version::V1_0,
         )
         .request_uri(state.presentation.base_url.clone())
@@ -1079,7 +1081,7 @@ mod tests {
     // See with: cargo test -- --nocapture client::tests::print_sizes
     #[test]
     fn print_sizes() {
-        for (name, size) in [
+        for (name, size) in &[
             ("RtspConnection", std::mem::size_of::<RtspConnection>()),
             (
                 "Session<Described>",
